import const
import note
import utility


class Chord(object):

    def __init__(self, note_list, duration, intensity, onset=0.0):

        self._note_list = None
        #: List. A list of note objects. You may alternately provide a list of
        #: strings in scientific pitch notation (e.g. ["C4", "E4", "G4"]) or a
        #: list of integer MIDI note numbers (e.g. [60, 64, 67]).
        self.note_list = note_list

        self._duration = None
        #: Float. The unit of measurement is abstract, so the user may
        #: interpret *duration* however they wish, e.g. absolute time, musical
        #: time, etc.
        self.duration = duration

        self._intensity = None
        #: String, as a dynamic mark (range "ppp" to "fff"). You may
        #: alternately provide an integer MIDI velocity (range 0 - 127).
        self.intensity = intensity

        self._onset = None
        #: Float. The time of onset. The unit of measurement is abstract, so
        #: the user may interpret *onset* however they wish, e.g. absolute
        #: time, musical time, etc.
        self.onset = onset

        # TODO: parent

    @property
    def note_list(self):
        return self._note_list

    @note_list.setter
    def note_list(self, value):
        if isinstance(value[0], basestring) or isinstance(value[0], int):
            l = []
            for v in value:
                l.append(note.Note(v, 0.0, 0))
            value = l
        self._note_list = value
        self.sort()

    @property
    def duration(self):
        return self._duration

    @duration.setter
    def duration(self, value):
        self._duration = value
        # unify duration
        for n in self.note_list:
            n.duration = self.duration

    @property
    def intensity(self):
        return self._intensity

    @intensity.setter
    def intensity(self, value):
        if isinstance(value, basestring):
            self._intensity = utility.dynamic_to_midi(str(value))
        else:
            self._intensity = value
        # unify intensity
        for n in self.note_list:
            n.intensity = self.intensity

    @property
    def onset(self):
        return self._onset

    @onset.setter
    def onset(self, value):
        self._onset = value
        # unify onset
        for n in self.note_list:
            n.onset = self.onset

    def __add__(self, item):
        # unify
        item.duration = self.duration
        item.intensity = self.intensity
        item.onset = self.onset
        # chord
        if isinstance(item, self.__class__):
            for n in item:
                self.note_list.append(n)
        # note
        if isinstance(item, note.Note):
            self.note_list.append(item)
        self.sort()

    def __contains__(self, item):
        return item in self.note_list

    def __delitem__(self, key):
        del self.note_list[key]

    def __getitem__(self, key):
        return self.note_list[key]

    def __iter__(self):
        return iter(self.note_list)

    def __len__(self):
        return len(self.note_list)

    def __repr__(self):
        s = ""
        for n in self.note_list:
            s += n.__repr__()
            s += "\n"
        return s

    def __setitem__(self, key, item):
        item.duration = self.duration
        item.intensity = self.intensity
        item.onset = self.onset
        self.note_list[key] = item
        self.sort()

    # info
    # ------------------------------------------------------------------- #

    def ambitus(self):
        """
        Return interval between highest and lowest Note objects in *note_list*.

        :returns: int -- interval

        """
        return self.limit(const.UPPER).pitch - self.limit(const.LOWER).pitch

    def limit(self, limit):
        """
        Return highest or lowest Note object as specified by *limit*, a
        predefined constant.

        :param limit: the limit to search (lower or upper)
        :type limit: int
        :returns: Note

        """
        result = self.note_list[0]
        for n in self.note_list:
            if limit == const.LOWER:
                if n.pitch < result.pitch:
                    result = n
            if limit == const.UPPER:
                if n.pitch > result.pitch:
                    result = n
        return result

    # round
    # ------------------------------------------------------------------- #

    def quantize(self):
        """
        Quantize each Note object in *note_list*.

        """
        for n in self.note_list:
            n.quantize()

    def tune(self):
        """
        Tune each Note object in *note_list*.

        """
        for n in self.note_list:
            n.tune()

    # sort
    # ------------------------------------------------------------------- #

    def sort(self):
        """
        Sort *note_list* by pitch.

        """
        self.note_list.sort(key=lambda x: x.pitch)

    # transform
    # ------------------------------------------------------------------- #

    def augment(self, factor):
        """
        Multiply *duration* by *factor* (i.e. scale on x axis). If *factor* is
        less than 1.0 but greater than 0.0, the effect will be diminution.

        :param factor: the factor of augmentation
        :type factor: float

        """
        self.duration *= factor

    def dilate(self, factor, reference=None):  # TODO: direction of dilation
        """
        Dilate a Chord object by *factor* (i.e. scale on y axis). The lowest
        Note object will remain constant unless a lower *reference* is
        specified. If *factor* is less than 1.0 but greater than 0.0, the
        effect will be contraction. If *factor* is negative, the dilation will
        be reflected.

        :param factor: the factor of dilation
        :type factor: float
        :param reference: an optional reference note for dilation
        :type reference: Note

        """
        lowest = self.limit(const.LOWER)
        if reference and reference.pitch < lowest.pitch:
            lowest = reference
        delta = (lowest.pitch * factor) - lowest.pitch
        for n in self.note_list:
            n.pitch *= factor
            n.pitch -= delta
        self.tune()

    def invert(self, note_):
        """
        Invert a Chord object about *note_* (i.e. reflect on x axis).

        :param note_: the axis of inversion
        :type note_: Note

        """
        for n in self.note_list:
            n.invert(note_)
        self.tune()
        self.sort()

    def shift(self, duration_):
        """
        Shift the time of onset by *duration_* (i.e. translate on x axis).

        :param duration_: a duration (positive or negative)
        :type duration_: float

        """
        self.onset += duration_

    def transpose(self, interval):
        """
        Transpose a Chord object by *interval* (i.e translate on y axis).

        :param interval: an interval in semitones (positive or negative)
        :type interval: int

        """
        for n in self.note_list:
            n.transpose(interval)
