import chord
import const
import note


class Fragment(object):

    def __init__(self, event_list, onset=0.0):

        #: List. A list of Note and/or Chord objects.
        self.event_list = event_list

        self._onset = None
        #: Float. The time of onset. The unit of measurement is abstract, so
        #: the user may interpret *onset* however they wish, e.g. absolute
        #: time, musical time, etc.
        self.onset = onset

    @property
    def onset(self):
        return self._onset

    @onset.setter
    def onset(self, value):
        self._onset = value
        self.queue()

    # TODO: magic methods

    def __repr__(self):
        s = "pit\t\tons\t\tdur\t\tdyn\n"
        for event in self.event_list:
            s += event.__repr__()
            if isinstance(event, note.Note):
                s += "\n"
            if event is not self.event_list[-1]:  # exclude last divider
                s += "---\n"
        return s

    # data
    # ------------------------------------------------------------------- #

    def onset_list(self):  # TODO: generalize
        result = []
        for event in self.event_list:
            if isinstance(event, chord.Chord):
                for n in event:
                    result.append(n.onset)
            if isinstance(event, note.Note):
                result.append(event.onset)
        return result

    def pitch_list(self):  # TODO: generalize
        result = []
        for event in self.event_list:
            if isinstance(event, chord.Chord):
                for n in event:
                    result.append(n.pitch)
            if isinstance(event, note.Note):
                result.append(event.pitch)
        return result

    # info
    # ------------------------------------------------------------------- #

    def ambitus(self):  # TODO: implement
        """
        Return interval between highest and lowest events in *event_list*.

        :returns: int -- interval

        """
        return self.limit(const.UPPER).pitch - self.limit(const.LOWER).pitch

    def duration(self):
        """
        Return duration of Fragment object.

        :returns: float -- duration

        """
        duration = 0.0
        for event in self.event_list:
            duration += event.duration
        return duration

    def limit(self, limit):
        """
        Return highest or lowest Note object as specified by *limit*, a
        predefined constant.

        :param limit: the limit to search (lower or upper)
        :type limit: int
        :returns: Note

        """
        result = None
        # separate chord and note
        chord_list = [x for x in self.event_list if isinstance(x, chord.Chord)]
        note_list = [x for x in self.event_list if isinstance(x, note.Note)]
        # extract limit from chord
        if chord_list:
            for c in chord_list:
                if limit == const.LOWER:
                    note_list.append(c.limit(const.LOWER))
                if limit == const.UPPER:
                    note_list.append(c.limit(const.UPPER))
        # find limit
        if note_list:
            result = note_list[0]
            for n in note_list:
                if limit == const.LOWER:
                    if n.pitch < result.pitch:
                        result = n
                if limit == const.UPPER:
                    if n.pitch > result.pitch:
                        result = n
        return result

    # round
    # ------------------------------------------------------------------- #

    def quantize(self):
        """
        Quantize event objects in *event_list*.

        """
        for event in self.event_list:
            event.quantize()

    def tune(self):
        """
        Tune event objects in *event_list*.

        """
        for event in self.event_list:
            event.tune()

    # sort
    # ------------------------------------------------------------------- #

    def queue(self):
        """
        Queue event objects in *event_list* relative to *onset*.

        """
        o = self.onset
        for event in self.event_list:
            event.onset = o
            o += event.duration

    def sort(self):
        """
        Sort *event_list* by *onset*.

        """
        self.event_list.sort(key=lambda x: x.onset)

    # transform
    # ------------------------------------------------------------------- #

    def augment(self, factor):
        """
        Multiply *duration* by *factor* (i.e. scale on x axis). If *factor* is
        less than 1.0 but greater than 0.0, the effect will be diminution.

        :param factor: the factor of augmentation
        :type factor: float

        """
        for event in self.event_list:
            event.augment(factor)
        self.queue()

    def dilate(self, factor):  # TODO: direction of dilation
        """
        Dilate a Fragment object by *factor* (i.e. scale on y axis). The lowest
        Note object will remain constant. If *factor* is less than 1.0 but
        greater than 0.0, the effect will be contraction. If *factor* is
        negative, the dilation will be reflected.

        :param factor: the factor of dilation
        :type factor: float

        """
        lowest = self.limit(const.LOWER)
        delta = (lowest.pitch * factor) - lowest.pitch
        for event in self.event_list:
            if isinstance(event, chord.Chord):
                event.dilate(factor, lowest)
            if isinstance(event, note.Note):
                event.pitch *= factor
                event.pitch -= delta
        self.tune()

    def invert(self, note_):
        """
        Invert a Fragment object about *note_* (i.e. reflect on x axis).

        :param note_: the axis of inversion
        :type note_: Note

        """
        for event in self.event_list:
            event.invert(note_)
        self.tune()

    def retrograde(self):
        """
        Reverse order of event objects in *event_list* (i.e. reflect on y axis).

        """
        for event in self.event_list:
            event.onset += event.duration
            event.onset *= -1
        self.sort()
        self.queue()

    def rotate(self, n):
        """
        Rotate *event_list* by *n* places. Positive values for *n* rotate
        right, whereas negative values for *n* rotate left.

        :param n: the number of places to rotate
        :type n: int

        """
        length = len(self.event_list)
        self.event_list = self.event_list[-n % length:] + \
            self.event_list[:-n % length]
        self.queue()

    def shift(self, duration):
        """
        Shift the time of onset by *duration* (i.e. translate on x axis).

        :param duration: a duration (positive or negative)
        :type duration: float

        """
        self.onset += duration

    def transpose(self, interval):
        """
        Transpose a Fragment object by *interval* (i.e translate on y axis).

        :param interval: an interval in semitones (positive or negative)
        :type interval: int

        """
        for event in self.event_list:
            event.transpose(interval)
