import math

import const
import fragment
import note
import section


class Interpolator(object):

    def __init__(self, origin, target, step_count, change):

        #: Fragment. The length of *origin.event_list* must match the length of
        #: *target.event_list*. This object is included in the output of
        #: *interpolate()*.
        self.origin = origin

        #: Fragment. The length of *target.event_list* must match the length of
        #: *origin.event_list*. This object is included in the output of
        #: *interpolate()*.
        self.target = target

        #: Integer. The number of intermediate Fragment objects to create.
        self.step_count = step_count

        #: Integer. The rate of change, or how quickly *target* begins to
        #: influence the intermediate Fragment objects. See Const Reference =>
        #: Rate of Change for supported rates of change.
        self.change = change

    def __repr__(self):
        return "%s %s %s %s" % (self.origin,
                                self.target,
                                self.step_count,
                                self.change)

    # interpolate
    # ------------------------------------------------------------------- #

    def interpolate(self):
        """
        Interpolate between *origin* and *target*, creating *step_count* number
        of intermediate Fragment objects at *change* rate of change. Note:
        this method can only interpolate between Note objects.

        :returns: Section

        """
        # first fragment is origin
        s = section.Section([self.origin], 0.0)
        # set up
        onset = s.duration()
        step = 1
        while step < self.step_count + 1:
            f = fragment.Fragment([], onset)
            # interpolate
            for i in range(len(self.origin.event_list)):
                o = self.origin.event_list[i]
                t = self.target.event_list[i]
                # calculate pitch, duration, intensity for current step
                pitch = self._value(o.pitch, t.pitch, step)
                duration = self._value(o.duration, t.duration, step)
                intensity = self._value(o.intensity, t.intensity, step)
                # create note
                n = note.Note(pitch, duration, intensity)
                # keep note
                f.event_list.append(n)
            # organize fragment
            f.tune()
            f.quantize()
            f.onset = onset
            # keep fragment
            s.fragment_list.append(f)
            # increment
            onset += f.duration()
            step += 1
        # last fragment is target
        self.target.onset = onset
        s.fragment_list.append(self.target)
        return s

    def _value(self, origin, target, step):
        delta = float(target) - float(origin)  # ensure float, avoid truncation
        increment = 0
        if self.change == const.LINEAR:
            # linear
            increment = delta / (self.step_count + 1)
            increment *= step
        if self.change == const.LOGARITHMIC:
            # logarithm (inverse)
            x = self.step_count + 1
            x -= step
            base = self.step_count + 1
            increment = math.log(x, base)
            increment = 1 - increment
            increment *= delta
        return origin + increment

# TODO: interpolate at different rates for pitch, duration, intensity
# TODO: increase variety of change, e.g. logarithmic (non-inverse), exponential
# TODO: interpolate between chord objects
