import math

import const
import utility


class Note(object):

    def __init__(self, pitch, duration, intensity, onset=0.0):  # TODO: parent

        self._pitch = None
        #: String, in scientific pitch notation (e.g. "C4"). You may
        #: alternately provide an integer MIDI note number (range 0 - 127).
        self.pitch = pitch

        #: Float. The unit of measurement is abstract, so the user may
        #: interpret *duration* however they wish, e.g. absolute time, musical
        #: time, etc.
        self.duration = duration

        self._intensity = None
        #: String, as a dynamic mark (range "ppp" to "fff"). You may
        #: alternately provide an integer MIDI velocity (range 0 - 127).
        self.intensity = intensity

        #: Float. The time of onset. The unit of measurement is abstract, so
        #: the user may interpret *onset* however they wish, e.g. absolute
        #: time, musical time, etc.
        self.onset = onset

    @property
    def pitch(self):
        return self._pitch

    @pitch.setter
    def pitch(self, value):
        if isinstance(value, basestring):
            self._pitch = self._scientific_to_midi(str(value))
        else:
            self._pitch = value

    @property
    def intensity(self):
        return self._intensity

    @intensity.setter
    def intensity(self, value):
        if isinstance(value, basestring):
            self._intensity = utility.dynamic_to_midi(str(value))
        else:
            self._intensity = value

    def __repr__(self):
        return "%s\t\t%s\t\t%s\t\t%s" % (self._midi_to_scientific(),
                                         self.onset,
                                         self.duration,
                                         self._midi_to_dynamic())

    # convert
    # ------------------------------------------------------------------- #

    def frequency(self):
        """
        Frequency in hertz (Hz) of receiving note.

        :returns: float -- Hz

        """
        interval = self.pitch - const.A4_MIDI
        f = const.A4_HZ * pow(2, float(interval) / const.PITCH_CLASS_COUNT)
        return round(f, 2)

    def _midi_to_dynamic(self):
        """
        Convert MIDI velocity to dynamic mark.

        :returns: str -- dynamic mark

        """
        return const.ID_MAP[self.intensity]  # TODO: round to nearest dynamic

    def _midi_to_scientific(self):
        """
        Convert MIDI note number to scientific pitch notation.

        :returns: str -- scientific pitch notation

        """
        pitch_class = const.PCN_MAP[self.pitch % const.PITCH_CLASS_COUNT]
        octave = self.pitch / const.PITCH_CLASS_COUNT
        octave = math.floor(octave)  # strip partial octave
        octave = int(octave)
        octave -= 1  # first octave is 0, not 1
        octave = str(octave)
        return pitch_class + octave

    @staticmethod
    def _scientific_to_midi(spn):
        """
        Convert scientific pitch notation to MIDI note number.

        :param spn: scientific pitch notation (e.g. "C4").
        :type spn: str
        :returns: int -- MIDI note number

        """
        pitch_class = None
        for key, value in const.PCN_MAP.items():
            if value == spn[:-1]:  # ignore last character
                pitch_class = key
                break
        octave = spn[-1]  # extract last character
        octave = int(octave)
        octave += 1  # first octave is 0, not 1
        return pitch_class + (octave * const.PITCH_CLASS_COUNT)

    # round
    # ------------------------------------------------------------------- #

    def quantize(self):
        """
        Round duration to a meaningful precision.

        """
        self.duration = round(self.duration, 2)

    def tune(self):
        """
        Round pitch to the nearest integer (i.e. MIDI note number).

        """
        self.pitch = int(round(self.pitch, 0))

    # transform
    # ------------------------------------------------------------------- #

    def augment(self, factor):
        """
        Multiply *duration* by *factor* (i.e. scale on x axis). If *factor* is
        less than 1.0 but greater than 0.0, the effect will be diminution.

        :param factor: the factor of augmentation
        :type factor: float

        """
        self.duration *= factor

    def invert(self, note):
        """
        Invert a Note object about *note* (i.e. reflect on x axis).

        :param note: the axis of inversion
        :type note: Note

        """
        self.pitch = note.pitch - (self.pitch - note.pitch)
        self.tune()

    def shift(self, duration_):
        """
        Shift the time of onset by *duration_* (i.e. translate on x axis).

        :param duration_: a duration (positive or negative)
        :type duration_: float

        """
        self.onset += duration_

    def transpose(self, interval):
        """
        Transpose a Note object by *interval* (i.e translate on y axis).

        :param interval: an interval in semitones (positive or negative)
        :type interval: int

        """
        self.pitch += interval
