import math

import chord
import const
import note


class RingModulator(object):

    def __init__(self, carrier, modulator, sidebands, ideal=True, diff=True):

        #: Note. Usually higher in pitch than *modulator*.
        self.carrier = carrier

        #: Note. Usually lower in pitch than *carrier*.
        self.modulator = modulator

        #: Integer. The number of sidebands to output.
        self.sidebands = sidebands

        #: Boolean. If False, output will include *carrier* and *modulator*.
        #: Default is True.
        self.ideal = ideal

        #: Boolean. If False, output will exclude difference tones. Default is
        #: True.
        self.diff = diff

    def __repr__(self):
        return "%s %s %s %s %s" % (self.carrier,
                                   self.modulator,
                                   self.sidebands,
                                   self.ideal,
                                   self.diff)

    # modulate
    # ------------------------------------------------------------------- #

    def modulate(self):
        """
        Create a spectrum via ring modulation.

        :returns: Chord

        """
        note_list = []
        i = 1
        # include carrier and modulator?
        if self.ideal is False:
            note_list.append(self.carrier)
            note_list.append(self.modulator)
        while i < self.sidebands + 1:
            carrier_freq = self.carrier.frequency()
            modulator_freq = self.modulator.frequency()
            # create upper sidebands
            freq = carrier_freq + (i * modulator_freq)
            n = note.Note(self._frequency_to_midi(freq),
                          self.carrier.duration,
                          self.carrier.intensity)
            note_list.append(n)
            # create lower sidebands
            if self.diff is True:
                freq = self._frequency_to_midi(carrier_freq -
                                              (i * modulator_freq))
                n = note.Note(self._frequency_to_midi(freq),
                              self.carrier.duration,
                              self.carrier.intensity)
                note_list.append(n)
            i += 1
        return chord.Chord(note_list,
                           self.carrier.duration,
                           self.carrier.intensity)

    # convert
    # ------------------------------------------------------------------- #

    @staticmethod
    def _frequency_to_midi(frequency):
        """
        Convert frequency to MIDI note number.

        :param frequency: frequency in hertz (Hz)
        :type frequency: float
        :returns: int -- MIDI note number

        """
        frequency /= const.A4_HZ
        frequency = abs(frequency)
        midi = (math.log(frequency, 2) * const.PITCH_CLASS_COUNT) + \
            const.A4_MIDI
        return int(round(midi))
